@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2024, The PurpleI2P Project
+* Copyright (c) 2013-2025, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -498,7 +498,8 @@ namespace garlic
 		buf += 4; // length
 
 		bool found = false;
-		if (SupportsEncryptionType (i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD))
+		bool supportsRatchets = SupportsRatchets ();
+		if (supportsRatchets)
 			// try ECIESx25519 tag
 			found = HandleECIESx25519TagMessage (buf, length);
 		if (!found)
@@ -535,7 +536,7 @@ namespace garlic
 					decryption->Decrypt(buf + 514, length - 514, iv, buf + 514);
 					HandleAESBlock (buf + 514, length - 514, decryption, msg->from);
 				}
-				else if (SupportsEncryptionType (i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD))
+				else if (supportsRatchets)
 				{
 					// otherwise ECIESx25519
 					auto ts = i2p::util::GetMillisecondsSinceEpoch ();
@@ -747,31 +748,35 @@ namespace garlic
 		std::shared_ptr<const i2p::data::RoutingDestination> destination, bool attachLeaseSet,
 	    bool requestNewIfNotFound)
 	{
-		if (destination->GetEncryptionType () == i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD &&
-			SupportsEncryptionType (i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD))
+		if (destination->GetEncryptionType () >= i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD)
 		{
-			ECIESX25519AEADRatchetSessionPtr session;
-			uint8_t staticKey[32];
-			destination->Encrypt (nullptr, staticKey); // we are supposed to get static key
-			auto it = m_ECIESx25519Sessions.find (staticKey);
-			if (it != m_ECIESx25519Sessions.end ())
-			{
-				session = it->second;
-				if (session->IsInactive (i2p::util::GetSecondsSinceEpoch ()))
+			if (SupportsEncryptionType (destination->GetEncryptionType ()))
+			{	
+				ECIESX25519AEADRatchetSessionPtr session;
+				uint8_t staticKey[32];
+				destination->Encrypt (nullptr, staticKey); // we are supposed to get static key
+				auto it = m_ECIESx25519Sessions.find (staticKey);
+				if (it != m_ECIESx25519Sessions.end ())
 				{
-					LogPrint (eLogDebug, "Garlic: Session restarted");
-					requestNewIfNotFound = true; // it's not a new session
-					session = nullptr;
+					session = it->second;
+					if (session->IsInactive (i2p::util::GetSecondsSinceEpoch ()))
+					{
+						LogPrint (eLogDebug, "Garlic: Session restarted");
+						requestNewIfNotFound = true; // it's not a new session
+						session = nullptr;
+					}
 				}
+				if (!session && requestNewIfNotFound)
+				{
+					session = std::make_shared<ECIESX25519AEADRatchetSession> (this, true);
+					session->SetRemoteStaticKey (destination->GetEncryptionType (), staticKey);
+				}
+				if (session && destination->IsDestination ())
+					session->SetDestination (destination->GetIdentHash ()); // NS or NSR
+				return session;
 			}
-			if (!session && requestNewIfNotFound)
-			{
-				session = std::make_shared<ECIESX25519AEADRatchetSession> (this, true);
-				session->SetRemoteStaticKey (staticKey);
-			}
-			if (session && destination->IsDestination ())
-				session->SetDestination (destination->GetIdentHash ()); // NS or NSR
-			return session;
+			else
+				LogPrint (eLogError, "Garlic: Non-supported encryption type ", destination->GetEncryptionType ());
 		}
 		else
 		{
@@ -999,7 +1004,8 @@ namespace garlic
 				i2p::fs::Remove (it);
 	}
 
-	void GarlicDestination::HandleECIESx25519GarlicClove (const uint8_t * buf, size_t len)
+	void GarlicDestination::HandleECIESx25519GarlicClove (const uint8_t * buf, size_t len,
+		ECIESX25519AEADRatchetSession * from)
 	{
 		const uint8_t * buf1 = buf;
 		uint8_t flag = buf[0]; buf++; // flag
@@ -1019,7 +1025,7 @@ namespace garlic
 				buf += 4; // expiration
 				ptrdiff_t offset = buf - buf1;
 				if (offset <= (int)len)
-					HandleCloveI2NPMessage (typeID, buf, len - offset, msgID);
+					HandleCloveI2NPMessage (typeID, buf, len - offset, msgID, from);
 				else
 					LogPrint (eLogError, "Garlic: Clove is too long");
 				break;
