@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2024, The PurpleI2P Project
+* Copyright (c) 2013-2025, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -14,6 +14,7 @@
 #include "Timestamp.h"
 #include "NetDb.hpp"
 #include "Tunnel.h"
+#include "CryptoKey.h"
 #include "LeaseSet.h"
 
 namespace i2p
@@ -399,6 +400,7 @@ namespace data
 		offset += propertiesLen; // skip for now. TODO: implement properties
 		// key sections
 		CryptoKeyType preferredKeyType = m_EncryptionType;
+		m_EncryptionType = 0;
 		bool preferredKeyFound = false;
 		if (offset + 1 > len) return 0;
 		int numKeySections = buf[offset]; offset++;
@@ -410,14 +412,22 @@ namespace data
 			if (offset + encryptionKeyLen > len) return 0;
 			if (IsStoreLeases () && !preferredKeyFound) // create encryptor with leases only
 			{
-				// we pick first valid key if preferred not found
-				auto encryptor = i2p::data::IdentityEx::CreateEncryptor (keyType, buf + offset);
-				if (encryptor && (!m_Encryptor || keyType == preferredKeyType))
-				{
-					m_Encryptor = encryptor; // TODO: atomic
-					m_EncryptionType = keyType;
-					if (keyType == preferredKeyType) preferredKeyFound = true;
-				}
+				// we pick max key type if preferred not found
+#if !OPENSSL_PQ
+				if (keyType <= i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD) // skip PQ keys if not supported
+#endif			
+				{	
+					if (keyType == preferredKeyType || !m_Encryptor || keyType > m_EncryptionType)
+					{
+						auto encryptor = i2p::data::IdentityEx::CreateEncryptor (keyType, buf + offset);
+						if (encryptor)
+						{
+							m_Encryptor = encryptor; // TODO: atomic
+							m_EncryptionType = keyType;
+							if (keyType == preferredKeyType) preferredKeyFound = true;
+						}	
+					}
+				}	
 			}
 			offset += encryptionKeyLen;
 		}
@@ -838,7 +848,7 @@ namespace data
 	}
 
 	LocalLeaseSet2::LocalLeaseSet2 (uint8_t storeType, const i2p::data::PrivateKeys& keys,
-		const KeySections& encryptionKeys, const std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> >& tunnels,
+		const EncryptionKeys& encryptionKeys, const std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> >& tunnels,
 		bool isPublic, uint64_t publishedTimestamp, bool isPublishedEncrypted):
 		LocalLeaseSet (keys.GetPublic (), nullptr, 0)
 	{
@@ -848,7 +858,7 @@ namespace data
 		if (num > MAX_NUM_LEASES) num = MAX_NUM_LEASES;
 		size_t keySectionsLen = 0;
 		for (const auto& it: encryptionKeys)
-			keySectionsLen += 2/*key type*/ + 2/*key len*/ + it.keyLen/*key*/;
+			keySectionsLen += 2/*key type*/ + 2/*key len*/ + it->pub.size()/*key*/;
 		m_BufferLen = identity->GetFullLen () + 4/*published*/ + 2/*expires*/ + 2/*flag*/ + 2/*properties len*/ +
 			1/*num keys*/ + keySectionsLen + 1/*num leases*/ + num*LEASE2_SIZE + keys.GetSignatureLen ();
 		uint16_t flags = 0;
@@ -883,9 +893,9 @@ namespace data
 		m_Buffer[offset] = encryptionKeys.size (); offset++; // 1 key
 		for (const auto& it: encryptionKeys)
 		{
-			htobe16buf (m_Buffer + offset, it.keyType); offset += 2; // key type
-			htobe16buf (m_Buffer + offset, it.keyLen); offset += 2; // key len
-			memcpy (m_Buffer + offset, it.encryptionPublicKey, it.keyLen); offset += it.keyLen; // key
+			htobe16buf (m_Buffer + offset, it->keyType); offset += 2; // key type
+			htobe16buf (m_Buffer + offset, it->pub.size()); offset += 2; // key len
+			memcpy (m_Buffer + offset, it->pub.data(), it->pub.size()); offset += it->pub.size(); // key
 		}
 		// leases
 		uint32_t expirationTime = 0; // in seconds
