@@ -688,7 +688,7 @@ namespace data
 				// since update was long time ago we assume that router is not connected anymore
 				r->ScheduleBufferToDelete ();
 			
-			if (r->GetProfile ()->IsUnreachable ())
+			if (r->HasProfile () && r->GetProfile ()->IsUnreachable ())
 				r->SetUnreachable (true);
 			// make router reachable back if too few routers or floodfills
 			if (r->IsUnreachable () && (total - deletedCount < NETDB_MIN_ROUTERS || isLowRate || isOffline ||
@@ -715,8 +715,9 @@ namespace data
 							r->SetUnreachable (true);
 				}	
 			}
-			// make router reachable back if connected now
-			if (r->IsUnreachable () && i2p::transport::transports.IsConnected (ident))
+			// make router reachable back if connected now or trusted router
+			if (r->IsUnreachable () && (i2p::transport::transports.IsConnected (ident) ||
+				i2p::transport::transports.IsTrustedRouter (ident)))
 				r->SetUnreachable (false);
 			
 			if (r->IsUnreachable ())
@@ -950,14 +951,13 @@ namespace data
 			LogPrint (eLogError, "NetDb: DatabaseLookup for zero ident. Ignored");
 			return;
 		}
-		char key[48];
-		int l = i2p::data::ByteStreamToBase64 (buf, 32, key, 48);
-		key[l] = 0;
+		std::string key;
+		if (CheckLogLevel (eLogInfo))
+			key = i2p::data::ByteStreamToBase64 (buf, 32);
 
 		IdentHash replyIdent(buf + 32);
 		uint8_t flag = buf[64];
 
-
 		LogPrint (eLogDebug, "NetDb: DatabaseLookup for ", key, " received flags=", (int)flag);
 		uint8_t lookupType = flag & DATABASE_LOOKUP_TYPE_FLAGS_MASK;
 		const uint8_t * excluded = buf + 65;
