@@ -29,13 +29,15 @@ namespace client
 	    const std::map<std::string, std::string>& params):
 		LeaseSetDestination (service, isPublic, &params),
 		m_Owner (owner), m_Identity (identity), m_EncryptionKeyType (m_Identity->GetCryptoKeyType ()),
-		m_IsCreatingLeaseSet (false), m_IsSameThread (isSameThread), m_LeaseSetCreationTimer (service)
+		m_IsCreatingLeaseSet (false), m_IsSameThread (isSameThread), 
+		m_LeaseSetCreationTimer (service), m_ReadinessCheckTimer (service)
 	{
 	}
 
 	void I2CPDestination::Stop ()
 	{
 		m_LeaseSetCreationTimer.cancel ();
+		m_ReadinessCheckTimer.cancel ();
 		LeaseSetDestination::Stop ();
 		m_Owner = nullptr;
 	}
@@ -77,7 +79,13 @@ namespace client
 		return keyType == i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD ? (bool)m_ECIESx25519Decryptor : m_EncryptionKeyType == keyType;
 	}
 
-
+	i2p::data::CryptoKeyType I2CPDestination::GetPreferredCryptoType () const
+	{
+		if (m_ECIESx25519Decryptor)
+			return i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD;
+		return i2p::data::CRYPTO_KEY_TYPE_ELGAMAL;
+	}	
+		
 	void I2CPDestination::HandleDataMessage (const uint8_t * buf, size_t len)
 	{
 		uint32_t length = bufbe32toh (buf);
@@ -88,7 +96,7 @@ namespace client
 
 	void I2CPDestination::CreateNewLeaseSet (const std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> >& tunnels)
 	{
-		boost::asio::post (GetService (), std::bind (&I2CPDestination::PostCreateNewLeaseSet, this, tunnels));
+		boost::asio::post (GetService (), std::bind (&I2CPDestination::PostCreateNewLeaseSet, GetSharedFromThis (), tunnels));
 	}
 
 	void I2CPDestination::PostCreateNewLeaseSet (std::vector<std::shared_ptr<i2p::tunnel::InboundTunnel> > tunnels)
@@ -98,6 +106,20 @@ namespace client
 			LogPrint (eLogInfo, "I2CP: LeaseSet is being created");
 			return;
 		}
+		m_ReadinessCheckTimer.cancel ();
+		auto pool = GetTunnelPool ();
+		if (!pool || pool->GetOutboundTunnels ().empty ())
+		{
+			// try again later
+			m_ReadinessCheckTimer.expires_from_now (boost::posix_time::seconds(I2CP_DESTINATION_READINESS_CHECK_INTERVAL));
+			m_ReadinessCheckTimer.async_wait(
+				[s=GetSharedFromThis (), tunnels=std::move(tunnels)](const boost::system::error_code& ecode)
+			    {
+					if (ecode != boost::asio::error::operation_aborted)
+						s->PostCreateNewLeaseSet (tunnels);
+				});	
+			return;
+		}	
 		uint8_t priv[256] = {0};
 		i2p::data::LocalLeaseSet ls (m_Identity, priv, tunnels); // we don't care about encryption key, we need leases only
 		m_LeaseSetExpirationTime = ls.GetExpirationTime ();
@@ -511,7 +533,7 @@ namespace client
 		if (sendBuf)
 		{
 			if (m_SendQueue.GetSize () < I2CP_MAX_SEND_QUEUE_SIZE)
-				m_SendQueue.Add (sendBuf);
+				m_SendQueue.Add (std::move(sendBuf));
 			else
 			{
 				LogPrint (eLogWarning, "I2CP: Send queue size exceeds ", I2CP_MAX_SEND_QUEUE_SIZE);
@@ -555,7 +577,7 @@ namespace client
 			m_IsSending = false;
 	}
 
-	std::string_view I2CPSession::ExtractString (const uint8_t * buf, size_t len)
+	std::string_view I2CPSession::ExtractString (const uint8_t * buf, size_t len) const
 	{
 		uint8_t l = buf[0];
 		if (l > len) l = len;
@@ -572,7 +594,7 @@ namespace client
 		return l + 1;
 	}
 
-	void I2CPSession::ExtractMapping (const uint8_t * buf, size_t len, std::map<std::string, std::string>& mapping)
+	void I2CPSession::ExtractMapping (const uint8_t * buf, size_t len, std::map<std::string, std::string>& mapping) const
 	// TODO: move to Base.cpp
 	{
 		size_t offset = 0;
@@ -1057,7 +1079,7 @@ namespace client
 		if (sendBuf)
 		{
 			if (m_SendQueue.GetSize () < I2CP_MAX_SEND_QUEUE_SIZE)
-				m_SendQueue.Add (sendBuf);
+				m_SendQueue.Add (std::move(sendBuf));
 			else
 			{
 				LogPrint (eLogWarning, "I2CP: Send queue size exceeds ", I2CP_MAX_SEND_QUEUE_SIZE);
