@@ -27,7 +27,7 @@ namespace client
 		m_Owner (owner), m_Socket(owner.GetService()), m_Timer (m_Owner.GetService ()),
 		m_BufferOffset (0),
 		m_SocketType (eSAMSocketTypeUnknown), m_IsSilent (false),
-		m_IsAccepting (false), m_Stream (nullptr)
+		m_IsAccepting (false), m_IsReceiving (false)
 	{
 	}
 
@@ -43,23 +43,21 @@ namespace client
 			m_Stream->AsyncClose ();
 			m_Stream = nullptr;
 		}
-		auto Session = m_Owner.FindSession(m_ID);
 		switch (m_SocketType)
 		{
 			case eSAMSocketTypeSession:
 				m_Owner.CloseSession (m_ID);
 			break;
 			case eSAMSocketTypeStream:
-			{
-				break;
-			}
+			break;
 			case eSAMSocketTypeAcceptor:
 			case eSAMSocketTypeForward:
 			{
-				if (Session)
+				auto session = m_Owner.FindSession(m_ID);
+				if (session)
 				{
-					if (m_IsAccepting && Session->GetLocalDestination ())
-						Session->GetLocalDestination ()->StopAcceptingStreams ();
+					if (m_IsAccepting && session->GetLocalDestination ())
+						session->GetLocalDestination ()->StopAcceptingStreams ();
 				}
 				break;
 			}
@@ -415,7 +413,7 @@ namespace client
 			{
 				session->UDPEndpoint = forward;
 				auto dest = session->GetLocalDestination ()->CreateDatagramDestination ();
-				auto port = std::stoi(params[SAM_PARAM_PORT]);
+				auto port = forward ? std::stoi(params[SAM_PARAM_PORT]) : 0;
 				if (type == eSAMSessionTypeDatagram)
 					dest->SetReceiver (std::bind (&SAMSocket::HandleI2PDatagramReceive, shared_from_this (),
 						std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4, std::placeholders::_5),
@@ -470,15 +468,11 @@ namespace client
 		auto session = m_Owner.FindSession(m_ID);
 		if (session)
 		{
-			uint8_t buf[1024];
-			char priv[1024];
-			size_t l = session->GetLocalDestination ()->GetPrivateKeys ().ToBuffer (buf, 1024);
-			size_t l1 = i2p::data::ByteStreamToBase64 (buf, l, priv, 1024);
-			priv[l1] = 0;
+			std::string priv = session->GetLocalDestination ()->GetPrivateKeys ().ToBase64 ();
 #ifdef _MSC_VER
-			size_t l2 = sprintf_s (m_Buffer, SAM_SOCKET_BUFFER_SIZE, SAM_SESSION_CREATE_REPLY_OK, priv);
+			size_t l2 = sprintf_s (m_Buffer, SAM_SOCKET_BUFFER_SIZE, SAM_SESSION_CREATE_REPLY_OK, priv.c_str ());
 #else
-			size_t l2 = snprintf (m_Buffer, SAM_SOCKET_BUFFER_SIZE, SAM_SESSION_CREATE_REPLY_OK, priv);
+			size_t l2 = snprintf (m_Buffer, SAM_SOCKET_BUFFER_SIZE, SAM_SESSION_CREATE_REPLY_OK, priv.c_str ());
 #endif
 			SendMessageReply (m_Buffer, l2, false);
 		}
@@ -643,53 +637,95 @@ namespace client
 			SendMessageReply (SAM_STREAM_STATUS_INVALID_ID, strlen(SAM_STREAM_STATUS_INVALID_ID), true);
 	}
 
-	void SAMSocket::ProcessStreamForward (char * buf, size_t len)
+	void SAMSocket::ProcessStreamForward(char* buf, size_t len)
 	{
-		LogPrint (eLogDebug, "SAM: Stream forward: ", buf);
+		LogPrint(eLogDebug, "SAM: Stream forward: ", buf);
+
 		std::map<std::string, std::string> params;
-		ExtractParams (buf, params);
-		std::string& id = params[SAM_PARAM_ID];
-		auto session = m_Owner.FindSession (id);
+		ExtractParams(buf, params);
+
+		const auto itId = params.find(SAM_PARAM_ID);
+		if (itId == params.end())
+		{
+			SendSessionI2PError("Missing ID");
+			return;
+		}
+		const std::string& id = itId->second;
+
+		auto session = m_Owner.FindSession(id);
 		if (!session)
 		{
-			SendMessageReply (SAM_STREAM_STATUS_INVALID_ID, strlen(SAM_STREAM_STATUS_INVALID_ID), true);
+			SendMessageReply(SAM_STREAM_STATUS_INVALID_ID, strlen(SAM_STREAM_STATUS_INVALID_ID), true);
 			return;
 		}
-		if (session->GetLocalDestination ()->IsAcceptingStreams ())
+		if (session->GetLocalDestination()->IsAcceptingStreams())
 		{
-			SendSessionI2PError ("Already accepting");
+			SendSessionI2PError("Already accepting");
 			return;
 		}
-		auto it = params.find (SAM_PARAM_PORT);
-		if (it == params.end ())
+
+		const auto itPort = params.find(SAM_PARAM_PORT);
+		if (itPort == params.end())
 		{
-			SendSessionI2PError ("PORT is missing");
+			SendSessionI2PError("PORT is missing");
 			return;
 		}
-		auto port = std::stoi (it->second);
+
+		const std::string& portStr = itPort->second;
+		if (!std::all_of(portStr.begin(), portStr.end(), ::isdigit))
+		{
+			SendSessionI2PError("Port must be numeric");
+			return;
+		}
+
+		int port = std::stoi(portStr);
 		if (port <= 0 || port >= 0xFFFF)
 		{
-			SendSessionI2PError ("Invalid PORT");
+			SendSessionI2PError("Invalid port");
 			return;
 		}
-		boost::system::error_code ec;
-		auto ep = m_Socket.remote_endpoint (ec);
-		if (ec)
+
+		boost::asio::ip::tcp::endpoint ep;
+		const auto itHost = params.find(SAM_PARAM_HOST);
+
+		if (itHost != params.end())
 		{
-			SendSessionI2PError ("Socket error");
-			return;
+			boost::system::error_code ec;
+			auto addr = boost::asio::ip::make_address(itHost->second, ec);
+			if (ec)
+			{
+				SendSessionI2PError("Invalid IP Address in HOST");
+				return;
+			}
+			ep = boost::asio::ip::tcp::endpoint(addr, port);
 		}
-		ep.port (port);
+		else
+		{
+			boost::system::error_code ec;
+			ep = m_Socket.remote_endpoint(ec);
+			if (ec)
+			{
+				SendSessionI2PError("Socket error: cannot get remote endpoint");
+				return;
+			}
+			ep.port(port);
+		}
+
 		m_SocketType = eSAMSocketTypeForward;
 		m_ID = id;
 		m_IsAccepting = true;
-		std::string& silent = params[SAM_PARAM_SILENT];
-		if (silent == SAM_VALUE_TRUE) m_IsSilent = true;
-		session->GetLocalDestination ()->AcceptStreams (std::bind (&SAMSocket::HandleI2PForward,
-			shared_from_this (), std::placeholders::_1, ep));
-		SendMessageReply (SAM_STREAM_STATUS_OK, strlen(SAM_STREAM_STATUS_OK), false);
+
+		auto itSilent = params.find(SAM_PARAM_SILENT);
+		if (itSilent != params.end() && itSilent->second == SAM_VALUE_TRUE)
+			m_IsSilent = true;
+
+		session->GetLocalDestination()->AcceptStreams(
+			std::bind(&SAMSocket::HandleI2PForward, shared_from_this(), std::placeholders::_1, ep));
+
+		SendMessageReply(SAM_STREAM_STATUS_OK, strlen(SAM_STREAM_STATUS_OK), false);
 	}
 
+
 	size_t SAMSocket::ProcessDatagramSend (char * buf, size_t len, const char * data)
 	{
 		LogPrint (eLogDebug, "SAM: Datagram send: ", buf, " ", len);
@@ -950,13 +986,29 @@ namespace client
 
 	void SAMSocket::Receive ()
 	{
-		m_Socket.async_read_some (boost::asio::buffer(m_Buffer + m_BufferOffset, SAM_SOCKET_BUFFER_SIZE - m_BufferOffset),
-			std::bind((m_SocketType == eSAMSocketTypeStream) ? &SAMSocket::HandleReceived : &SAMSocket::HandleMessage,
-			shared_from_this (), std::placeholders::_1, std::placeholders::_2));
+		if (m_SocketType == eSAMSocketTypeStream)
+		{
+			if (m_IsReceiving) return;
+			size_t bufSize = SAM_SOCKET_BUFFER_SIZE;
+			size_t unsentSize = m_Stream ? m_Stream->GetSendBufferSize () : 0;
+			if (unsentSize)
+			{	
+				if (unsentSize >= SAM_STREAM_MAX_SEND_BUFFER_SIZE) return; // buffer is full
+				if (unsentSize > SAM_STREAM_MAX_SEND_BUFFER_SIZE - SAM_SOCKET_BUFFER_SIZE)
+					bufSize = SAM_STREAM_MAX_SEND_BUFFER_SIZE - unsentSize;
+			}
+			m_IsReceiving = true;
+			m_Socket.async_read_some (boost::asio::buffer(m_Buffer, bufSize),
+				std::bind(&SAMSocket::HandleReceived, shared_from_this (), std::placeholders::_1, std::placeholders::_2));
+		}	
+		else
+			m_Socket.async_read_some (boost::asio::buffer(m_Buffer + m_BufferOffset, SAM_SOCKET_BUFFER_SIZE - m_BufferOffset),
+				std::bind(&SAMSocket::HandleMessage, shared_from_this (), std::placeholders::_1, std::placeholders::_2));
 	}
 
 	void SAMSocket::HandleReceived (const boost::system::error_code& ecode, std::size_t bytes_transferred)
 	{
+		m_IsReceiving = false;
 		if (ecode)
 		{
 			LogPrint (eLogError, "SAM: Read error: ", ecode.message ());
@@ -966,16 +1018,13 @@ namespace client
 		else
 		{
 			if (m_Stream)
-			{
-				bytes_transferred += m_BufferOffset;
-				m_BufferOffset = 0;
+			{	
 				m_Stream->AsyncSend ((uint8_t *)m_Buffer, bytes_transferred,
 					std::bind(&SAMSocket::HandleStreamSend, shared_from_this(), std::placeholders::_1));
-			}
+				Receive ();
+			}	
 			else
-			{
 				Terminate("No Stream Remaining");
-			}
 		}
 	}
 
@@ -986,16 +1035,16 @@ namespace client
 			if (m_Stream->GetStatus () == i2p::stream::eStreamStatusNew ||
 				m_Stream->GetStatus () == i2p::stream::eStreamStatusOpen) // regular
 			{
-				m_Stream->AsyncReceive (boost::asio::buffer (m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE),
+				m_Stream->AsyncReceive (boost::asio::buffer (m_StreamBuffer, SAM_STREAM_BUFFER_SIZE),
 						std::bind (&SAMSocket::HandleI2PReceive, shared_from_this(),
 						std::placeholders::_1, std::placeholders::_2),
 							SAM_SOCKET_CONNECTION_MAX_IDLE);
 			}
 			else // closed by peer
 			{
-				uint8_t * buff = new uint8_t[SAM_SOCKET_BUFFER_SIZE];
+				uint8_t * buff = new uint8_t[SAM_STREAM_BUFFER_SIZE];
 				// get remaining data
-				auto len = m_Stream->ReadSome (buff, SAM_SOCKET_BUFFER_SIZE);
+				auto len = m_Stream->ReadSome (buff, SAM_STREAM_BUFFER_SIZE);
 				if (len > 0) // still some data
 				{
 					WriteI2PDataImmediate(buff, len);
@@ -1116,18 +1165,19 @@ namespace client
 				}
 			}
 			if (!m_IsSilent)
-			{
-				// get remote peer address
-				auto ident_ptr = stream->GetRemoteIdentity();
-				const size_t ident_len = ident_ptr->GetFullLen();
-				uint8_t* ident = new uint8_t[ident_len];
-
-				// send remote peer address as base64
-				const size_t l = ident_ptr->ToBuffer (ident, ident_len);
-				const size_t l1 = i2p::data::ByteStreamToBase64 (ident, l, (char *)m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE);
-				delete[] ident;
-				m_StreamBuffer[l1] = '\n';
-				HandleI2PReceive (boost::system::error_code (), l1 +1); // we send identity like it has been received from stream
+			{			
+				if (m_SocketType != eSAMSocketTypeTerminated)
+				{
+					// get remote peer address
+					auto ident = std::make_shared<std::string>(stream->GetRemoteIdentity()->ToBase64 ()); // we need to keep it until sent
+					ident->push_back ('\n');
+					// send remote peer address back to client like received from stream
+					boost::asio::async_write (m_Socket, boost::asio::buffer (ident->data (), ident->size ()), boost::asio::transfer_all(),
+						[ident, s = shared_from_this ()](const boost::system::error_code& ecode, size_t bytes_transferred)
+						{
+							s->HandleWriteI2PData (ecode, bytes_transferred);
+						});
+				}	
 			}
 			else
 				I2PReceive ();
@@ -1191,11 +1241,11 @@ namespace client
 			else
 			{
 #ifdef _MSC_VER
-				size_t l = sprintf_s ((char *)m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE, SAM_DATAGRAM_RECEIVED, base64.c_str (), (long unsigned int)len);
+				size_t l = sprintf_s ((char *)m_StreamBuffer, SAM_STREAM_BUFFER_SIZE, SAM_DATAGRAM_RECEIVED, base64.c_str (), (long unsigned int)len);
 #else
-				size_t l = snprintf ((char *)m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE, SAM_DATAGRAM_RECEIVED, base64.c_str (), (long unsigned int)len);
+				size_t l = snprintf ((char *)m_StreamBuffer, SAM_STREAM_BUFFER_SIZE, SAM_DATAGRAM_RECEIVED, base64.c_str (), (long unsigned int)len);
 #endif
-				if (len < SAM_SOCKET_BUFFER_SIZE - l)
+				if (len < SAM_STREAM_BUFFER_SIZE - l)
 				{
 					memcpy (m_StreamBuffer + l, buf, len);
 					WriteI2PData(len + l);
@@ -1219,11 +1269,11 @@ namespace client
 			else
 			{
 #ifdef _MSC_VER
-				size_t l = sprintf_s ((char *)m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE, SAM_RAW_RECEIVED, (long unsigned int)len);
+				size_t l = sprintf_s ((char *)m_StreamBuffer, SAM_STREAM_BUFFER_SIZE, SAM_RAW_RECEIVED, (long unsigned int)len);
 #else
-				size_t l = snprintf ((char *)m_StreamBuffer, SAM_SOCKET_BUFFER_SIZE, SAM_RAW_RECEIVED, (long unsigned int)len);
+				size_t l = snprintf ((char *)m_StreamBuffer, SAM_STREAM_BUFFER_SIZE, SAM_RAW_RECEIVED, (long unsigned int)len);
 #endif
-				if (len < SAM_SOCKET_BUFFER_SIZE - l)
+				if (len < SAM_STREAM_BUFFER_SIZE - l)
 				{
 					memcpy (m_StreamBuffer + l, buf, len);
 					WriteI2PData(len + l);
@@ -1482,17 +1532,39 @@ namespace client
 			session->StopLocalDestination ();
 			session->Close ();
 			if (m_IsSingleThread)
-			{
-				auto timer = std::make_shared<boost::asio::deadline_timer>(GetService ());
-				timer->expires_from_now (boost::posix_time::seconds(5)); // postpone destination clean for 5 seconds
-				timer->async_wait ([timer, session](const boost::system::error_code& ecode)
-				{
-					// session's destructor is called here
-				});
-			}
+				ScheduleSessionCleanupTimer (session); // let all session's streams close
 		}
 	}
 
+	void SAMBridge::ScheduleSessionCleanupTimer (std::shared_ptr<SAMSession> session)
+	{
+		auto timer = std::make_shared<boost::asio::deadline_timer>(GetService ());
+		timer->expires_from_now (boost::posix_time::seconds(5)); // postpone destination clean for 5 seconds
+		timer->async_wait (std::bind (&SAMBridge::HandleSessionCleanupTimer, this, std::placeholders::_1, session, timer));
+	}	
+		
+	void SAMBridge::HandleSessionCleanupTimer (const boost::system::error_code& ecode,
+		std::shared_ptr<SAMSession> session, std::shared_ptr<boost::asio::deadline_timer> timer)
+	{
+		if (ecode != boost::asio::error::operation_aborted && session)
+		{
+			auto dest = session->GetLocalDestination ();
+			if (dest)
+			{
+				auto streamingDest = dest->GetStreamingDestination ();
+				auto numStreams = streamingDest->GetNumStreams ();
+				if (numStreams > 0)
+				{
+					LogPrint (eLogInfo, "SAM: Session ", session->Name, " still has ", numStreams, " streams");
+					ScheduleSessionCleanupTimer (session);
+				}	
+				else
+					LogPrint (eLogDebug, "SAM: Session ", session->Name, " terminated");
+			}	
+		}	
+		// session's destructor is called here unless rescheduled
+	}	
+		
 	std::shared_ptr<SAMSession> SAMBridge::FindSession (const std::string& id) const
 	{
 		std::unique_lock<std::mutex> l(m_SessionsMutex);
